@@ -185,7 +185,7 @@ g_array_new_take (gpointer  data,
 
   g_return_val_if_fail (data != NULL || len == 0, NULL);
   g_return_val_if_fail (len <= G_MAXUINT, NULL);
-  g_return_val_if_fail (element_size <= G_MAXUINT, NULL);
+  g_return_val_if_fail (element_size > 0 && element_size <= G_MAXUINT, NULL);
 
   array = g_array_sized_new (FALSE, clear, element_size, 0);
   rarray = (GRealArray *) array;
@@ -198,7 +198,9 @@ g_array_new_take (gpointer  data,
 
 /**
  * g_array_new_take_zero_terminated: (skip)
- * @data: (array zero-terminated=1): an array of elements of @element_size
+ * @data: (array zero-terminated=1) (transfer full) (nullable): an array
+ *     of elements of @element_size, %NULL terminated,
+ *     or %NULL for an empty array
  * @clear: %TRUE if #GArray elements should be automatically cleared
  *     to 0 when they are allocated
  * @element_size: the size of each element in bytes
@@ -232,10 +234,11 @@ g_array_new_take_zero_terminated (gpointer  data,
                                   gboolean  clear,
                                   gsize     element_size)
 {
+  GRealArray *rarray;
   GArray *array;
   gsize len = 0;
 
-  g_return_val_if_fail (element_size <= G_MAXUINT, NULL);
+  g_return_val_if_fail (element_size > 0 && element_size <= G_MAXUINT, NULL);
 
   if (data != NULL)
     {
@@ -253,10 +256,13 @@ g_array_new_take_zero_terminated (gpointer  data,
         }
     }
 
-  g_return_val_if_fail (len <= G_MAXUINT, NULL);
+  g_return_val_if_fail (len < G_MAXUINT, NULL);
 
   array = g_array_new_take (data, len, clear, element_size);
-  ((GRealArray *)array)->zero_terminated = TRUE;
+  rarray = (GRealArray *) array;
+  rarray->zero_terminated = TRUE;
+  if (G_LIKELY (rarray->data != NULL))
+    rarray->elt_capacity = len + 1;
 
   return array;
 }
@@ -271,8 +277,9 @@ g_array_new_take_zero_terminated (gpointer  data,
  * the underlying array is preserved for use elsewhere and returned
  * to the caller.
  *
- * If the array was created with the @zero_terminate property
- * set to %TRUE, the returned data is zero terminated too.
+ * Note that if the array was created with the @zero_terminate
+ * property set to %TRUE, this may still return %NULL if the length
+ * of the array was zero and data was not yet allocated.
  *
  * If array elements contain dynamically-allocated memory,
  * the array elements should also be freed by the caller.
@@ -770,11 +777,12 @@ g_array_set_size (GArray *farray,
     }
   else if (length < array->len)
     g_array_remove_range (farray, length, array->len - length);
-  
+
   array->len = length;
-  
-  g_array_zero_terminate (array);
-  
+
+  if (G_LIKELY (array->data != NULL))
+    g_array_zero_terminate (array);
+
   return farray;
 }
 
@@ -881,6 +889,9 @@ g_array_remove_range (GArray *farray,
   g_return_val_if_fail (index_ <= G_MAXUINT - length, NULL);
   g_return_val_if_fail (index_ + length <= array->len, NULL);
 
+  if (length == 0)
+    return farray;
+
   if (array->clear_func != NULL)
     {
       guint i;
@@ -892,7 +903,7 @@ g_array_remove_range (GArray *farray,
   if (index_ + length != array->len)
     memmove (g_array_elt_pos (array, index_),
              g_array_elt_pos (array, index_ + length),
-             (array->len - (index_ + length)) * array->elt_size);
+             g_array_elt_len (array, array->len - (index_ + length)));
 
   array->len -= length;
   if (G_UNLIKELY (g_mem_gc_friendly))
@@ -1026,7 +1037,7 @@ g_array_binary_search (GArray        *array,
         {
           middle = left + (right - left) / 2;
 
-          val = compare_func (_array->data + (_array->elt_size * middle), target);
+          val = compare_func (g_array_elt_pos (_array, middle), target);
           if (val == 0)
             {
               result = TRUE;
@@ -1143,10 +1154,6 @@ ptr_array_new (guint reserved_size,
 
   if (reserved_size != 0)
     {
-      if (G_LIKELY (reserved_size < G_MAXUINT) &&
-          null_terminated)
-        reserved_size++;
-
       g_ptr_array_maybe_expand (array, reserved_size);
       g_assert (array->pdata != NULL);
 
@@ -1258,6 +1265,7 @@ GPtrArray *
 g_ptr_array_new_take_null_terminated (gpointer       *data,
                                       GDestroyNotify  element_free_func)
 {
+  GRealPtrArray *rarray;
   GPtrArray *array;
   gsize len = 0;
 
@@ -1267,10 +1275,13 @@ g_ptr_array_new_take_null_terminated (gpointer       *data,
         len += 1;
     }
 
-  g_return_val_if_fail (len <= G_MAXUINT, NULL);
+  g_return_val_if_fail (len < G_MAXUINT, NULL);
 
   array = g_ptr_array_new_take (g_steal_pointer (&data), len, element_free_func);
-  ((GRealPtrArray *)array)->null_terminated = TRUE;
+  rarray = (GRealPtrArray *) array;
+  rarray->null_terminated = TRUE;
+  if (G_LIKELY (rarray->pdata != NULL))
+    rarray->alloc = len + 1;
 
   return array;
 }
@@ -1287,7 +1298,7 @@ ptr_array_new_from_array (gpointer       *data,
   GRealPtrArray *rarray;
 
   g_assert (data != NULL || len == 0);
-  g_assert (len <= G_MAXUINT);
+  g_assert (len <= G_MAXUINT - (null_terminated ? 1 : 0));
 
   array = ptr_array_new (len, element_free_func, null_terminated);
   rarray = (GRealPtrArray *)array;
@@ -1398,7 +1409,7 @@ g_ptr_array_new_from_null_terminated_array (gpointer       *data,
     }
 
   g_assert (data != NULL || len == 0);
-  g_return_val_if_fail (len <= G_MAXUINT, NULL);
+  g_return_val_if_fail (len < G_MAXUINT, NULL);
 
   return ptr_array_new_from_array (
     data, len, copy_func, copy_func_user_data, element_free_func, TRUE);
@@ -1523,7 +1534,7 @@ g_ptr_array_copy (GPtrArray *array,
 
   if (rarray->alloc > 0)
     {
-      g_ptr_array_maybe_expand ((GRealPtrArray *) new_array, array->len + rarray->null_terminated);
+      g_ptr_array_maybe_expand ((GRealPtrArray *) new_array, array->len);
 
       if (array->len > 0)
         {
@@ -1586,11 +1597,11 @@ g_array_copy (GArray *array)
   g_return_val_if_fail (rarray != NULL, NULL);
 
   new_rarray =
-    (GRealArray *) g_array_sized_new (rarray->zero_terminated, rarray->clear,
-                                      rarray->elt_size, rarray->elt_capacity);
+      (GRealArray *) g_array_sized_new (rarray->zero_terminated, rarray->clear,
+                                        rarray->elt_size, rarray->len);
   new_rarray->len = rarray->len;
   if (rarray->len > 0)
-    memcpy (new_rarray->data, rarray->data, rarray->len * rarray->elt_size);
+    memcpy (new_rarray->data, rarray->data, g_array_elt_len (rarray, rarray->len));
 
   g_array_zero_terminate (new_rarray);
 
@@ -1878,22 +1889,24 @@ static void
 g_ptr_array_maybe_expand (GRealPtrArray *array,
                           guint          len)
 {
-  guint max_len;
+  guint max_len, want_len;
 
   /* The maximum array length is derived from following constraints:
    * - The number of bytes must fit into a gsize / 2.
    * - The number of elements must fit into guint.
+   * - null terminated arrays must leave space for the terminating element
    */
-  max_len = MIN (G_MAXSIZE / 2 / sizeof (gpointer), G_MAXUINT);
+  max_len = MIN (G_MAXSIZE / 2 / sizeof (gpointer), G_MAXUINT) - (array->null_terminated ? 1 : 0);
 
   /* Detect potential overflow */
   if G_UNLIKELY ((max_len - array->len) < len)
     g_error ("adding %u to array would overflow", len);
 
-  if ((array->len + len) > array->alloc)
+  want_len = array->len + len + (array->null_terminated ? 1 : 0);
+  if (want_len > array->alloc)
     {
       guint old_alloc = array->alloc;
-      gsize want_alloc = g_nearest_pow (sizeof (gpointer) * (array->len + len));
+      gsize want_alloc = g_nearest_pow (sizeof (gpointer) * want_len);
       want_alloc = MAX (want_alloc, MIN_ARRAY_SIZE);
       array->alloc = MIN (want_alloc / sizeof (gpointer), G_MAXUINT);
       array->pdata = g_realloc (array->pdata, want_alloc);
@@ -1930,11 +1943,7 @@ g_ptr_array_set_size  (GPtrArray *array,
     {
       guint i;
 
-      if (G_UNLIKELY (rarray->null_terminated) &&
-          length_unsigned - rarray->len > G_MAXUINT - 1)
-        g_error ("array would overflow");
-
-      g_ptr_array_maybe_expand (rarray, (length_unsigned - rarray->len) + rarray->null_terminated);
+      g_ptr_array_maybe_expand (rarray, length_unsigned - rarray->len);
 
       /* This is not
        *     memset (array->pdata + array->len, 0,
@@ -2219,7 +2228,7 @@ g_ptr_array_add (GPtrArray *array,
   g_return_if_fail (rarray);
   g_return_if_fail (rarray->len == 0 || (rarray->len != 0 && rarray->pdata != NULL));
 
-  g_ptr_array_maybe_expand (rarray, 1u + rarray->null_terminated);
+  g_ptr_array_maybe_expand (rarray, 1u);
 
   rarray->pdata[rarray->len++] = data;
 
@@ -2264,11 +2273,7 @@ g_ptr_array_extend (GPtrArray  *array_to_extend,
   if (array->len == 0u)
     return;
 
-  if (G_UNLIKELY (array->len == G_MAXUINT) &&
-      rarray_to_extend->null_terminated)
-    g_error ("adding %u to array would overflow", array->len);
-
-  g_ptr_array_maybe_expand (rarray_to_extend, array->len + rarray_to_extend->null_terminated);
+  g_ptr_array_maybe_expand (rarray_to_extend, array->len);
 
   if (func != NULL)
     {
@@ -2345,7 +2350,7 @@ g_ptr_array_insert (GPtrArray *array,
   g_return_if_fail (index_ >= -1);
   g_return_if_fail (index_ < 0 || (guint) index_ <= rarray->len);
 
-  g_ptr_array_maybe_expand (rarray, 1u + rarray->null_terminated);
+  g_ptr_array_maybe_expand (rarray, 1u);
 
   real_index = (index_ >= 0) ? (guint) index_ : rarray->len;
 
